// A trivial (but nevertheless, working quite well) nonogram solver,
// based on binary decision diagrams (BDD).
//
// For more information on nonograms (or Paint-by-Number Puzzle) see e.g.
// http://webpbn.com/
// 
// The numbered programs are sequential parts of presentation, showing 
// different stages of solver development. The first fully working nonogram
// solver is in nonogrambdd2.cpp.
//
// This particular file demonstrates generation of variants of a single
// row/column, using weak integer compositions.
// 
// The program requires BuDDy (libbdd-dev package on Debian). Supposing BuDDy
// is installed system-wide, compilation can be done with the command: 
// g++ -o nonogrambdd1 -lbdd -O2 nonogrambdd1.cpp
// or using Makefile, provided with the package.
//                                                                          
// (c) 2011 Konstantin L. Metlov <metlov@fti.dn.ua>                         
//                                                                          
// You can redistribute and/or modify this file under the terms of the GNU  
// General Public License as published by the Free Software Foundation, eith
// version 3 of the License, or (at your option) any later version.         
//                                                                          
// This file is distributed in the hope that it will be useful,             
// but WITHOUT ANY WARRANTY; without even the implied warranty of           
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            
// GNU General Public License for more details.

#include <cstdlib>                                                          
#include <cmath>                                                            
#include <cassert>
#include <iostream>
#include <iomanip>
#include <bdd.h>

using namespace std;

void comp_first(int n, int k, int* x);
bool comp_next(int n, int k, int* x);

// Each nonogram is stored sequentially in a single array: first the number
// of rows, then the number of columns, then all rows, then all columns.
// Information for each row/column consists of, first, number
// of colored dot blocks m, followed by m positive numbers, representing
// block lengths.

// given a position of row/column in the nonogram array, return the n of the
// corresponding sequence of integer decompositions. nr is the total number 
// of elements in row/column.
int nonogram_n(int nr, int pos, int* nonogram) {
  int s=0;
  int m=nonogram[pos];
  for (int j=0; j<m; j++) {
    s+=nonogram[pos+1+j];
    if (j<m-1) s++; // count mandatory white space between blocks
  };
  return nr-s;
};

// given a position of row/column in the nonogram array, return the k of the
// corresponding sequence of integer decompositions.
int nonogram_k(int nr, int pos, int* nonogram) {
  return nonogram[pos]+1; // before all blocks, after all blocks, and in-between
};

// convert an integer decomposition x into the true/false bits in the array,
// representing a particular state of row/column. Array bits must have nr 
// elements.
void nonogram_comptobits(int pos, int* nonogram, int n, int k, int* x, 
			int nr, bool* bits) {
  int cb=0;
  for(int i=0;i<x[0];i++) bits[cb++]=false;
  int m=nonogram[pos];
  for (int j=0;j<m;j++) {
    for(int i=0;i<nonogram[pos+1+j];i++) bits[cb++]=true;
    if (j<m-1) bits[cb++]=false; // mandatory white space
    for(int i=0;i<x[j+1];i++) bits[cb++]=false;
  };
  assert(cb==nr);
};


int main(int argc, char *argv[]) { 
  // // print out some nonogram row/column variants
  int nonogram[] = {3,1,2,4};
  const int nr=11;
  bool bits[nr];
  int n=nonogram_n(nr, 0, nonogram);
  int k=nonogram_k(nr, 0, nonogram);
  int x[k];
  comp_first(n,k,x);
  do {
    for (int i=0;i<k;i++) cout << setw(2) << x[i] << ' ';
    cout << endl;

    nonogram_comptobits(0, nonogram, n, k, x, nr, bits);
    for (int i=0;i<nr;i++)
      if (bits[i]) cout << '*'; else cout << '-';
    cout << endl;
    
    cout << endl;
  } while (comp_next(n,k,x));
};

// initializes compisition generator to enumerate compositions of
// n into k parts with the first composition returned in the array x
// (must be of the length k).
void comp_first(int n, int k, int* x) {
  x[0]=n;
  for(int i=1;i<k;i++) x[i]=0; 
};

// generates the next composition into the array x. returns false if the
// composition in the array x is already lexicographically the last one.
bool comp_next(int n, int k, int* x) {
  int s=0;
  for (int i=0; i<k; i++) {
    assert(x[i]>=0);
    s+=x[i];
  };
  assert(s==n);
  if (x[k-1]==n) return false;
  int i=0;
  while(x[i]==0) i++; // find the first non-zero element
  int v=x[i];
  x[i]=0;
  x[0]=v-1;
  x[i+1]++;
  return true;
};
