// A trivial (but nevertheless, working quite well) nonogram solver,
// based on binary decision diagrams (BDD).
//
// For more information on nonograms (or Paint-by-Number Puzzle) see e.g.
// http://webpbn.com/
// 
// The numbered programs are sequential parts of presentation, showing 
// different stages of solver development. The first fully working nonogram
// solver is in nonogrambdd2.cpp.
//
// This particular file contains a larger example, compared to nonogrambdd2,
// which, nevertheless, is easily solved by the program. The solver itself
// is the same.
// 
// The program requires BuDDy (libbdd-dev package on Debian). Supposing BuDDy
// is installed system-wide, compilation can be done with the command: 
// g++ -o nonogrambdd3 -lbdd -O2 nonogrambdd3.cpp
// or using Makefile, provided with the package.
//                                                                          
// (c) 2011 Konstantin L. Metlov <metlov@fti.dn.ua>                         
//                                                                          
// You can redistribute and/or modify this file under the terms of the GNU  
// General Public License as published by the Free Software Foundation, eith
// version 3 of the License, or (at your option) any later version.         
//                                                                          
// This file is distributed in the hope that it will be useful,             
// but WITHOUT ANY WARRANTY; without even the implied warranty of           
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            
// GNU General Public License for more details.

#include <cstdlib>                                                          
#include <cmath>                                                            
#include <cassert>
#include <iostream>
#include <iomanip>
#include <bdd.h>

using namespace std;

void comp_first(int n, int k, int* x);
bool comp_next(int n, int k, int* x);

// Each nonogram is stored sequentially in a single array: first the number
// of rows, then the number of columns, then all rows, then all columns.
// Information for each row/column consists of, first, number
// of colored dot blocks m, followed by m positive numbers, representing
// block lengths.

// given a position of row/column in the nonogram array, return the n of the
// corresponding sequence of integer decompositions. nr is the total number 
// of elements in row/column.
int nonogram_n(int nr, int pos, int* nonogram) {
  int s=0;
  int m=nonogram[pos];
  for (int j=0; j<m; j++) {
    s+=nonogram[pos+1+j];
    if (j<m-1) s++; // count mandatory white space between blocks
  };
  return nr-s;
};

// given a position of row/column in the nonogram array, return the k of the
// corresponding sequence of integer decompositions.
int nonogram_k(int nr, int pos, int* nonogram) {
  return nonogram[pos]+1; // before all blocks, after all blocks, and in-between
};

// convert an integer decomposition x into the true/false bits in the array,
// representing a particular state of row/column. Array bits must have nr 
// elements.
void nonogram_comptobits(int pos, int* nonogram, int n, int k, int* x, 
			int nr, bool* bits) {
  int cb=0;
  for(int i=0;i<x[0];i++) bits[cb++]=false;
  int m=nonogram[pos];
  for (int j=0;j<m;j++) {
    for(int i=0;i<nonogram[pos+1+j];i++) bits[cb++]=true;
    if (j<m-1) bits[cb++]=false; // mandatory white space
    for(int i=0;i<x[j+1];i++) bits[cb++]=false;
  };
  assert(cb==nr);
};

// bits in the nonogram are the bdd variables, numbered linearly
// this function returns global indices of variables, representing
// the elements of a particular i-th row. The array ri must have space
// for nc elements;
void nonogram_row_indices(int nr, int nc, int ir, int* ri) {
  for(int i=0;i<nc;i++) {
    ri[i]=ir*nc+i; // row-major ordering
    assert(ri[i]<nr*nc);
  };
};

void nonogram_col_indices(int nc, int nr, int ic, int* ci) {
  for(int i=0;i<nr;i++) {
    ci[i]=i*nc+ic; // row-major ordering
    assert(ci[i]<nr*nc);
  };
};

double tot_variants=1.0;

// appends constraints, representing all possible block assignments for
// row/column. Rows and columns distinguished here by passing different
// functions for mapping them (indexed by a number) to bdd variables.
void add_constraints(bdd &nonobdd, int &pos, int* nonogram, int nr, int nc, 
		     void (*indices)(int, int, int, int *)) {
  for (int ir=0; ir<nr; ir++) {
    int nvariants=0;
    bool bits[nc];
    int n=nonogram_n(nc, pos, nonogram);
    int k=nonogram_k(nc, pos, nonogram);
    int ri[nc];
    (*indices)(nr, nc, ir, ri);
    
    int x[k];
    comp_first(n,k,x);
    
    bdd rowbdd = bdd_false();
    do {
      nonogram_comptobits(pos, nonogram, n, k, x, nc, bits);
      bdd variantbdd = bdd_true();
      for (int i=0; i<nc; i++) {
	bdd bit=bits[i]?bddtruepp:bddfalsepp;
	variantbdd&=bdd_biimp(bit, bdd_ithvarpp(ri[i]));
      };
      rowbdd|=variantbdd;
      nvariants++;
    } while (comp_next(n,k,x));
    nonobdd &= rowbdd;

    pos+=(nonogram[pos]+1); // move next
    tot_variants*=nvariants;
  };
};

static int nr;
static int nc;
void nonoPrint(char* varset, int size);

int hiusednodeswatermark=0;                                                 
int lousednodeswatermark=0;                                                 
void my_gbchandler(int pre, bddGbcStat *s);

int main(int argc, char *argv[]) { 
  // "Rock-n-roll..." from
  // http://www.picross.co.uk/HardPicrossPuzzles.aspx?puzzle=5
  int nonogram[]= { 20, 20, // 20 x 20
  		    1, 2,   // first row
  		    2, 4, 1,
  		    3, 2, 1, 1,
  		    4, 4, 1, 1, 2,
  		    3, 2, 1, 2,
  		    2, 2, 2,
  		    1, 2,
  		    1, 2,
  		    1, 6,
  		    3, 5, 2, 8,
  		    5, 6, 1, 4, 1, 1,
  		    2, 8, 8,
  		    5, 6, 1, 1, 1, 2,
  		    5, 8, 2, 1, 1, 1,
  		    6, 4, 4, 1, 1, 1, 2,
  		    6, 3, 3, 2, 1, 1, 1,
  		    6, 3, 3, 1, 1, 1, 2,
  		    6, 4, 4, 2, 1, 1, 1,
  		    2, 8, 8,
  		    2, 6, 11,
  		    1, 4, // first column
  		    2, 3, 6,
  		    1,12,
  		    4, 1, 1, 7, 3,
  		    2,14, 2,
  		    2,14, 2,
  		    5, 1, 1, 1, 5, 3,
  		    2, 2, 9,
  		    2, 3, 6,
  		    2, 4, 1,
  		    1, 1,
  		    2, 1,11,
  		    5, 3, 3, 1, 1, 3,
  		    4, 4, 1, 1, 2,
  		    5, 1, 3, 1, 1, 3,
  		    6, 3, 1, 2, 1, 1, 2,
  		    4, 3, 1, 1, 3,
  		    6, 1, 1, 2, 1, 1, 2,
  		    2, 3,11,
  		    1, 1
  };
  
  nr = nonogram[0];
  nc = nonogram[1];
  int pos= 2; // position of the first row
  
  bdd_init(8000000,1000000);
  bdd_setmaxincrease(0);
  bdd_gbc_hook(my_gbchandler); // monitor garbage collection

  bdd_setvarnum(nr*nc); // number of variables
  
  bdd nonobdd = bdd_true(); // assume we have the solution

  // append constraints for all rows
  add_constraints(nonobdd, pos, nonogram, nr, nc, nonogram_row_indices);
  // append constraints for all columns
  add_constraints(nonobdd, pos, nonogram, nc, nr, nonogram_col_indices);
  // voila ! nonobdd contains all the solutions
  cout << "There are " << bdd_satcount(nonobdd) << " solutions." << endl;
  // print them
  cout << "Here they are:" << endl;
  bdd_allsat(nonobdd,nonoPrint);

  cout << tot_variants << " variants considered." << endl;

  bdd_gbc();
  cout << "Nodespace soft peak usage "                                      
       << (double(hiusednodeswatermark)/bdd_getallocnum()*100) << "% ("     
       << hiusednodeswatermark << " out of "                                
       << bdd_getallocnum() << ")."<<endl;                                  
  cout << "Nodespace hard peak usage "                                      
       << (double(lousednodeswatermark)/bdd_getallocnum()*100) << "% ("     
       << lousednodeswatermark << " out of "                                
       << bdd_getallocnum() << ")."<<endl;
};

// initializes compisition generator to enumerate compositions of
// n into k parts with the first composition returned in the array x
// (must be of the length k).
void comp_first(int n, int k, int* x) {
  x[0]=n;
  for(int i=1;i<k;i++) x[i]=0; 
};

// generates the next composition into the array x. returns false if the
// composition in the array x is already lexicographically the last one.
bool comp_next(int n, int k, int* x) {
  int s=0;
  for (int i=0; i<k; i++) {
    assert(x[i]>=0);
    s+=x[i];
  };
  assert(s==n);
  if (x[k-1]==n) return false;
  int i=0;
  while(x[i]==0) i++; // find the first non-zero element
  int v=x[i];
  x[i]=0;
  x[0]=v-1;
  x[i+1]++;
  return true;
};

// print the nonogram nicely
void nonoPrint(char* varset, int size) {
  for (int i=0; i<nr; i++) {
    for (int j=0; j<nc; j++) {
      int v=i*nc+j;
      cout << (varset[v] < 0 ? "??" : (varset[v] > 0 ? "[]" : "  "));
    };
    cout << endl;
  };
  cout << endl;
};


// garbage collection callback to collect statistics                        
void my_gbchandler(int pre, bddGbcStat *s) {                                                                           
  int used=(s->nodes)-(s->freenodes);                                       
  if (pre) {                                                                
    if (used>hiusednodeswatermark) hiusednodeswatermark=used;               
  } else {                                                                  
    if (used>lousednodeswatermark) lousednodeswatermark=used;               
  }                                                                         
                                                                            
  if (!pre) {                                                               
    int nodes=s->nodes;                                                     
    int softpct=int(double(hiusednodeswatermark)/nodes*100);                
    int hardpct=int(double(lousednodeswatermark)/nodes*100);                
    printf("GC #%d: %d nodes / %d free (usage: %d%% soft, %d%% hard)",      
           s->num, s->nodes, s->freenodes, softpct, hardpct);               
    printf(" [ %.1fs / %.1fs ]\n",                                          
           (float)s->time/(float)(CLOCKS_PER_SEC),                          
           (float)s->sumtime/(float)CLOCKS_PER_SEC);                        
  }                                                                         
}
